@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2005-2012  NTT DATA CORPORATION
  *
- * Version: 1.8.3+   2011/12/13
+ * Version: 1.8.3+   2015/04/21
  */
 
 #ifndef _LINUX_CCSECURITY_H
@@ -29,32 +29,37 @@ struct sock;
 struct sk_buff;
 struct msghdr;
 struct pid_namespace;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
+int search_binary_handler(struct linux_binprm *bprm);
+#else
 int search_binary_handler(struct linux_binprm *bprm, struct pt_regs *regs);
+#endif
 
 #ifdef CONFIG_CCSECURITY
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
-/* Obtain prototype of __d_path() or d_absolute_path(). */
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36) && LINUX_VERSION_CODE < KERNEL_VERSION(3, 2, 0)
+/* Obtain prototype of __d_path(). */
 #include <linux/dcache.h>
 #endif
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
-/* Obtain prototype of find_task_by_vpid() and find_task_by_pid_ns(). */
-#include <linux/sched.h>
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+/* Obtain definition of kuid_t and kgid_t. */
+#include <linux/uidgid.h>
 #endif
 
 /* For exporting variables and functions. */
 struct ccsecurity_exports {
 	void (*load_policy) (const char *filename);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 2, 0)
-	typeof(d_absolute_path) (*d_absolute_path);
+	char * (*d_absolute_path) (const struct path *, char *, int);
 #elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
 	typeof(__d_path) (*__d_path);
 #elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
 	spinlock_t *vfsmount_lock;
 #endif
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
-	typeof(find_task_by_vpid) (*find_task_by_vpid);
-	typeof(find_task_by_pid_ns) (*find_task_by_pid_ns);
+	struct task_struct * (*find_task_by_vpid) (pid_t nr);
+	struct task_struct * (*find_task_by_pid_ns) (pid_t nr,
+						     struct pid_namespace *ns);
 #endif
 };
 
@@ -65,14 +70,14 @@ struct ccsecurity_operations {
 	int (*chroot_permission) (struct path *path);
 	int (*pivot_root_permission) (struct path *old_path,
 				      struct path *new_path);
-	int (*mount_permission) (char *dev_name, struct path *path,
+	int (*mount_permission) (const char *dev_name, struct path *path,
 				 const char *type, unsigned long flags,
 				 void *data_page);
 #else
 	int (*chroot_permission) (struct nameidata *nd);
 	int (*pivot_root_permission) (struct nameidata *old_nd,
 				      struct nameidata *new_nd);
-	int (*mount_permission) (char *dev_name, struct nameidata *nd,
+	int (*mount_permission) (const char *dev_name, struct nameidata *nd,
 				 const char *type, unsigned long flags,
 				 void *data_page);
 #endif
@@ -130,16 +135,25 @@ struct ccsecurity_operations {
 					  struct msghdr *msg, int size);
 	int (*socket_post_recvmsg_permission) (struct sock *sk,
 					       struct sk_buff *skb, int flags);
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+	int (*chown_permission) (struct dentry *dentry, struct vfsmount *mnt,
+				 kuid_t user, kgid_t group);
+#else
 	int (*chown_permission) (struct dentry *dentry, struct vfsmount *mnt,
 				 uid_t user, gid_t group);
+#endif
 	int (*chmod_permission) (struct dentry *dentry, struct vfsmount *mnt,
 				 mode_t mode);
 	int (*getattr_permission) (struct vfsmount *mnt,
 				   struct dentry *dentry);
 	int (*sigqueue_permission) (pid_t pid, int sig);
 	int (*tgsigqueue_permission) (pid_t tgid, pid_t pid, int sig);
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
+	int (*search_binary_handler) (struct linux_binprm *bprm);
+#else
 	int (*search_binary_handler) (struct linux_binprm *bprm,
 				      struct pt_regs *regs);
+#endif
 #ifdef CONFIG_CCSECURITY_USE_EXTERNAL_TASK_SECURITY
 	int (*alloc_task_security) (const struct task_struct *task);
 	void (*free_task_security) (const struct task_struct *task);
@@ -165,14 +179,13 @@ static inline int ccs_pivot_root_permission(struct path *old_path,
 	return func ? func(old_path, new_path) : 0;
 }
 
-static inline int ccs_mount_permission(char *dev_name, struct path *path,
-				       char *type, unsigned long flags,
+static inline int ccs_mount_permission(const char *dev_name, struct path *path,
+				       const char *type, unsigned long flags,
 				       void *data_page)
 {
-	int (*func) (char *, struct path *, const char *, unsigned long,
+	int (*func) (const char *, struct path *, const char *, unsigned long,
 		     void *) = ccsecurity_ops.mount_permission;
-	return func ? func(dev_name, path, (const char *) type, flags,
-			   data_page) : 0;
+	return func ? func(dev_name, path, type, flags, data_page) : 0;
 }
 
 #else
@@ -191,14 +204,13 @@ static inline int ccs_pivot_root_permission(struct nameidata *old_nd,
 	return func ? func(old_nd, new_nd) : 0;
 }
 
-static inline int ccs_mount_permission(char *dev_name, struct nameidata *nd,
-				       char *type, unsigned long flags,
-				       void *data_page)
+static inline int ccs_mount_permission(const char *dev_name,
+				       struct nameidata *nd, const char *type,
+				       unsigned long flags, void *data_page)
 {
-	int (*func) (char *, struct nameidata *, const char *, unsigned long,
-		     void *) = ccsecurity_ops.mount_permission;
-	return func ? func(dev_name, nd, (const char *) type, flags,
-			   data_page) : 0;
+	int (*func) (const char *, struct nameidata *, const char *,
+		     unsigned long, void *) = ccsecurity_ops.mount_permission;
+	return func ? func(dev_name, nd, type, flags, data_page) : 0;
 }
 
 #endif
@@ -357,6 +369,19 @@ static inline int ccs_uselib_permission(struct dentry *dentry,
 
 #endif
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+
+static inline int ccs_chown_permission(struct dentry *dentry,
+				       struct vfsmount *mnt, kuid_t user,
+				       kgid_t group)
+{
+	int (*func) (struct dentry *, struct vfsmount *, kuid_t, kgid_t)
+		= ccsecurity_ops.chown_permission;
+	return func ? func(dentry, mnt, user, group) : 0;
+}
+
+#else
+
 static inline int ccs_chown_permission(struct dentry *dentry,
 				       struct vfsmount *mnt, uid_t user,
 				       gid_t group)
@@ -366,6 +391,8 @@ static inline int ccs_chown_permission(struct dentry *dentry,
 	return func ? func(dentry, mnt, user, group) : 0;
 }
 
+#endif
+
 static inline int ccs_chmod_permission(struct dentry *dentry,
 				       struct vfsmount *mnt, mode_t mode)
 {
@@ -374,12 +401,23 @@ static inline int ccs_chmod_permission(struct dentry *dentry,
 	return func ? func(dentry, mnt, mode) : 0;
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
+
+static inline int ccs_search_binary_handler(struct linux_binprm *bprm)
+{
+	return ccsecurity_ops.search_binary_handler(bprm);
+}
+
+#else
+
 static inline int ccs_search_binary_handler(struct linux_binprm *bprm,
 					    struct pt_regs *regs)
 {
 	return ccsecurity_ops.search_binary_handler(bprm, regs);
 }
 
+#endif
+
 #else
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 25)
@@ -395,8 +433,8 @@ static inline int ccs_pivot_root_permission(struct path *old_path,
 	return 0;
 }
 
-static inline int ccs_mount_permission(char *dev_name, struct path *path,
-				       char *type, unsigned long flags,
+static inline int ccs_mount_permission(const char *dev_name, struct path *path,
+				       const char *type, unsigned long flags,
 				       void *data_page)
 {
 	return 0;
@@ -415,9 +453,9 @@ static inline int ccs_pivot_root_permission(struct nameidata *old_nd,
 	return 0;
 }
 
-static inline int ccs_mount_permission(char *dev_name, struct nameidata *nd,
-				       char *type, unsigned long flags,
-				       void *data_page)
+static inline int ccs_mount_permission(const char *dev_name,
+				       struct nameidata *nd, const char *type,
+				       unsigned long flags, void *data_page)
 {
 	return 0;
 }
@@ -537,6 +575,17 @@ static inline int ccs_fcntl_permission(struct file *file, unsigned int cmd,
 	return 0;
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
+
+static inline int ccs_chown_permission(struct dentry *dentry,
+				       struct vfsmount *mnt, kuid_t user,
+				       kgid_t group)
+{
+	return 0;
+}
+
+#else
+
 static inline int ccs_chown_permission(struct dentry *dentry,
 				       struct vfsmount *mnt, uid_t user,
 				       gid_t group)
@@ -544,12 +593,23 @@ static inline int ccs_chown_permission(struct dentry *dentry,
 	return 0;
 }
 
+#endif
+
 static inline int ccs_chmod_permission(struct dentry *dentry,
 				       struct vfsmount *mnt, mode_t mode)
 {
 	return 0;
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
+
+static inline int ccs_search_binary_handler(struct linux_binprm *bprm)
+{
+	return search_binary_handler(bprm);
+}
+
+#else
+
 static inline int ccs_search_binary_handler(struct linux_binprm *bprm,
 					    struct pt_regs *regs)
 {
@@ -558,6 +618,8 @@ static inline int ccs_search_binary_handler(struct linux_binprm *bprm,
 
 #endif
 
+#endif
+
 #ifdef CONFIG_CCSECURITY_USE_EXTERNAL_TASK_SECURITY
 
 static inline int ccs_alloc_task_security(const struct task_struct *task)
